class FiForm::Builder # PORO \o/

  attr_reader :resource, :form, :options, :indentation
  attr_accessor :validation_errors

  def initialize(resource, options={}, template, &block)
    @resource = resource
    @template = template
    @options = options.dup

    @validation_errors = {}

    @form = build_form
    @target = @form
    @indentation = options[:indentation] || 0
    @renderer = options.delete(:renderer)
  end

  alias_method :object, :resource

  def add(kind, options={})
    @target << build(kind, options)
    self # allow chaining
  end

  def group(title=nil, &block)
    add(:group, title: title, indentation: indentation).add_children do
      @template.capture(self, &block)
    end
  end

  def input(field, options={})
    input_options = {
      value: options[:value] || resource_value(field),
      field: field,
      label: options[:label],
    }

    input_options[:hint] = options[:hint] if options[:hint]

    html_options = options.except(:as, :label, :hint, :value)
    input_options[:input] = html_options if html_options.any?

    add(:input, input_options)
  end

  def render(*args)
    content = @template.capture(self) do
      @template.render *args
    end
    add :raw, content: content
  end

  def hint(hint, options={})
    add(:hint, options.merge(hint: hint, indentation: indentation))
  end

  def title(title, options={})
    add(:title, options.merge(title: title, indentation: indentation))
  end

  def submit(label=nil, options={})
    add(:submit, options.merge(label: label, indentation: indentation))
  end

  def to_html
    FiForm.configuration.get_renderer(@renderer).new(@template).render(@form)
  end

  def add_children(indent: true, &block)
    with_target last_item, indent: indent, &block
  end

  def self.add_builder_method(name, &block)
    define_method name, &block
  end

private
  def resource_value(field)
  end

  def build(kind, options)
    FiForm::Item.build(kind, form: @form, **options)
  end

  def sub_builder(resource, &block)
    FiForm::Builder.new(resource, @options.merge(indentation: indentation), @template, &block)
  end

  def last_item
    @target.last_child
  end

  def with_target(target_list, indent: true, &block)
    previous_target = @target
    @indentation += 1 if indent
    @target = target_list
    yield
    @indentation -= 1 if indent
    @target = previous_target
  end

  def capture_target(options={}, &block)
    indent = !!options.delete(:indent)

    target_list = []
    with_target(target_list, indent: indent, &block)

    single_target = target_list.first
    single_target.options.merge! options
    single_target
  end

  def build_form
    FiForm::Item.build(:form,
      resource: @resource,
      errors: validation_errors,
      form_options: @options
    )
  end
end